// Return values rules.
// Except when differently indicated:
//
// int-returning functions:
//   Values >= 0 are returned values (not errors), while
//   values < 0 are errors
//
// char*-returing functions:
//   !NULL is normal return value
//   NULL is error (no specific indication of which error occurred)

#include <time.h>

#ifndef ABCANON_INCLUDE
#define ABCANON_INCLUDE

#ifndef ABCAPI
 #define ABCAPI __stdcall
#endif

#define ABC_RETVAL_OK                  0
#define ABC_RETVAL_ERROR             (-1)    // Generic
#define ABC_RETVAL_BUSY              (-2)
#define ABC_RETVAL_BUFFER_TOO_SMALL  (-3)
#define ABC_RETVAL_INVALID_HANDLE    (-4)
#define ABC_RETVAL_INVALID_PARM      (-5)
#define ABC_RETVAL_OUT_OF_MEMORY     (-6)
#define ABC_RETVAL_NOT_FOUND         (-7)
#define ABC_RETVAL_SEM_ERROR         (-8)
#define ABC_RETVAL_INVALID_PATH      (-9)
#define ABC_RETVAL_TIMEOUT          (-10)
#define ABC_RETVAL_DENIED           (-11)
#define ABC_RETVAL_NOT_CONNECTED    (-12)
#define ABC_RETVAL_THREAD_BUSY      (-13)
// Other errors (that should be encoded as 0x80xxxxxx) are Canon error xxxxxx

#define ABC_ERROR(a)        ((a)<0)

typedef void *  ABCAMERA_HANDLE;

void ABCAPI ABCanonGetVersion (uint32_t *pVersionH, uint32_t *pVersionL, uint32_t *pBuild);

// Scan for supported cameras and return the number of cameras found
struct sCameraData
{
 uint32_t                   Idx;

 #define ABC_STS_OK                   0
 #define ABC_STS_INUSE                1
 #define ABC_STS_CREATEFILE           2
 #define ABC_STS_WINUSBINIT           3
 #define ABC_STS_QUERYDEVINFO         4
 #define ABC_STS_EPNUMBER             5
 #define ABC_STS_EPTYPE_0             6
 #define ABC_STS_EPTYPE_1             7
 #define ABC_STS_EPTYPE_2             8
 #define ABC_STS_CONTROLRW            9
 #define ABC_STS_INITSTATUS          10
 #define ABC_STS_BADSIGNATURE        11
 #define ABC_STS_UNKNOWNCAMERA       12
 #define ABC_STS_CMD12CHECKSFAIL     13
 #define ABC_STS_UNKNOWNERROR        14
 #define ABC_STS_CMD2512CHECKSFAIL   15
 #define ABC_STS_SETPIPEPOLICY       16
 #define ABC_STS_
 uint32_t                   Status;

 uint32_t                   SerialNumber;
 unsigned char              CameraName[32];

 // From here on, byte alignment
 uint8_t                    FWRelease[3];

 #define ABCTYPE_UNKNOWN        0x00
 #define ABCTYPE_350D           0x01
 #define ABCTYPE_5D             0x02
 uint8_t                    CameraType;
};

int ABCAPI ABCanonBuildCameraList (struct sCameraData **ppCameraData);
void ABCAPI ABCanonReleaseCameraList (struct sCameraData *pCameraData);

// ConnectMethod:
#define CONNECT_FIRST_AVAILABLE              0    // Parm unused
#define CONNECT_BY_INDEX                     1    // Parm = Index
#define CONNECT_BY_SERIAL_NUMBER             2    // Parm = Serial Number
// Filter:
#define ABCTYPE_ALL                          0    // And also all the ABCTYPEs defined above
int ABCAPI ABCanonConnectToCamera (int ConnectMethod, int Filter, uint32_t Parm, ABCAMERA_HANDLE *phCH);
int ABCAPI ABCanonDisconnectFromCamera (ABCAMERA_HANDLE hCH);

// Raw notification routed directly from Canon interrupt notification
typedef void _stdcall (*ABC_RAW_NOTIFICATION_CB)(ABCAMERA_HANDLE Handle, uint16_t CanonEventID, uint16_t Severity, uint8_t *pData, int DataSize);
int ABCAPI ABCanonRegisterRawNotificationCB (ABCAMERA_HANDLE hCH, ABC_RAW_NOTIFICATION_CB lNotificationFunction);

// ABCanon library "translated" event.
// Some events can be discarded by the notification routine by returning !0.
// Some events when not discarded will be queued and can be retrieved with ABCanonGetEvent
struct sABCEvent
{
 uint32_t     EventID;
 uint32_t     EventSize;
 void        *EventPtr;
};
// Event can be discarded (ignored) ------------------------+  by returning !0 from the NotificationCB
// Event will be enqueued even when discarded ------------+ |
// Event will never be enqueued ------------------------+ | |
#define ABC_EVENT_NONE                           0   // | | |
#define ABC_EVENT_BATTERY_STATUS_CHANGED         1   // - - *  EventSize is the new battery status
#define ABC_EVENT_SETTINGS_CHANGED               2   // - - *  EventPtr is NULL
#define ABC_EVENT_SHOT_COMPLETED                 3   // * - -  EventSize is camera counter. Just an informational event, occurs only once regardless of the picture quality setting,
                                                     //                 **AFTER** the ImgCounter variable has been incremented (so you can change it if you want)
#define ABC_EVENT_JPG_READY                      4   // * - -  EventSize is camera counter. If return value is 0, the image(s) will be discarded!
#define ABC_EVENT_CR2_READY                      5   // * - -  EventSize is camera counter. If return value is 0, the image(s) will be discarded!
#define ABC_EVENT_DOWNLOADING                    6   // * - -  EventPtr is NULL, EventSize is % (0 to 100), done twice in case of RAW+JPG. With ABC_OPTIONS_PCT2 EventPtr
                                                     //                 is a pointer to a struct sDownloadPct variable holding other information.
#define ABC_EVENT_JPG_SAVED                      7   // - - *  EventPtr is the filename (WSTR) saved. EventSize is the file length (DownloadMode == ABC_IMGDL_SAVE_TO_FILE)
#define ABC_EVENT_CR2_SAVED                      8   // - - *  EventPtr is the filename (WSTR) saved. EventSize is the file length (DownloadMode == ABC_IMGDL_SAVE_TO_FILE)
#define ABC_EVENT_JPG_SAVE_FAILED                9   // - - *  EventPtr is the image data, EventSize is the image size (DownloadMode == ABC_IMGDL_SAVE_TO_FILE)
#define ABC_EVENT_CR2_SAVE_FAILED               10   // - - *  EventPtr is the image data, EventSize is the image size (DownloadMode == ABC_IMGDL_SAVE_TO_FILE)
#define ABC_EVENT_JPG_IMAGE                     11   // - - *  EventPtr is the image data, EventSize is the image size (DownloadMode == ABC_IMGDL_SAVE_TO_MEMORY)
#define ABC_EVENT_CR2_IMAGE                     12   // - - *  EventPtr is the image data, EventSize is the image size (DownloadMode == ABC_IMGDL_SAVE_TO_MEMORY)
#define ABC_EVENT_AF_FAILED                     13   // - - *
#define ABC_EVENT_CAMERA_DISCONNECTED           14   // - - *
#define ABC_EVENT_JPG_DISCARDED                 15   // - - *
#define ABC_EVENT_CR2_DISCARDED                 16   // - - *
#define ABC_EVENT_BUSY                          17   // - - *  2.03: EventPtr is NULL, EventSize is 1 (Busy) or 0 (no longer BUSY). Requires ABC_OPTIONS_BUSYPOLL
#define ABC_EVENT_ALL_DONE                      18   // - - *  2.03: EventPtr is NULL, EventSize is 0.
typedef int _stdcall (*ABC_NOTIFICATION_CB)(ABCAMERA_HANDLE Handle, struct sABCEvent *pABCEvent);
int ABCAPI ABCanonRegisterNotificationCB (ABCAMERA_HANDLE hCH, ABC_NOTIFICATION_CB lNotificationFunction);

struct sDownloadPct
{
 uint32_t    TotalImages;            // Total images to download for this shot: 1 or 2 in case of RAW+JPG
 uint32_t    CurrentImage;           // 0/1. The current image
 uint32_t    FullPct;                // The 0-100% percentage of the full download. Does always 0-100% regardless of the number of images to download.
};

// If Timeout==INFINITE return when an event is available.
// For other timeout values, return value is ABC_RETVAL_TIMEOUT if no event gets queued within Timeout milliseconds
int ABCAPI ABCanonGetEvent (ABCAMERA_HANDLE hCH, struct sABCEvent *pEvent, uint32_t Timeout);
// Event memory should be released calling the ABCanonReleaseEvent function
void ABCAPI ABCanonReleaseEvent (struct sABCEvent *pEvent);


// Image settings
#define ABC_IMGSETTING_MODE_DIAL                0        // Read-only
#define ABC_IMGSETTING_SHUTTER_SPEED            1
#define ABC_IMGSETTING_APERTURE                 2
#define ABC_IMGSETTING_ISO                      3
#define ABC_IMGSETTING_EXP_COMPENSATION         4
#define ABC_IMGSETTING_METERING                 5
#define ABC_IMGSETTING_AUTOFOCUS                6        // Read-only
#define ABC_IMGSETTING_WHITE_BALANCE            7
#define ABC_IMGSETTING_PICTURE_QUALITY          8
#define ABC_IMGSETTING_FLASH_EXP_COMP           9
#define ABC_IMGSETTING_AUTO_POWER_OFF          10        // Read-only
#define ABC_IMGSETTING_DRIVE_MODE              11        // Read-only
#define ABC_IMGSETTING_AEB                     12        // Read-only
#define ABC_IMGSETTING_COLOR_TEMPERATURE       13        // 5D only
#define ABC_IMGSETTING_

int ABCAPI ABCanonGetImgSetting (ABCAMERA_HANDLE hCH, int ImgSetting);

// Return values is an error or the number of used entries of pParmArray
int ABCAPI ABCanonGetAllSettings (ABCAMERA_HANDLE hCH, int32_t *pSettingsArray, int SettingsArraySize);

char * ABCAPI ABCanonGetImgSettingText (ABCAMERA_HANDLE hCH, int ImgSetting);
char * ABCAPI ABCanonImgSettingValueToText (int ImgSetting, int Value);

#define ABC_SETTYPE_FIRST          0   // Lowest value, Value is ignored
#define ABC_SETTYPE_CHANGE         1   // Relative change (with sign)
#define ABC_SETTYPE_SET            2   // Set to the Value parameter
#define ABC_SETTYPE_LAST           3   // Set to highest value, Value is ignored
#define ABC_SETTYPE_PREV           4   // Switch to previous. Same as CHANGE -1, but wraps from BOL to EOL. Value is ignored
#define ABC_SETTYPE_NEXT           5   // Switch to next. Same as change +1, but wraps from EOL to BOL. Value is ignored
int ABCAPI ABCanonSetImgSetting (ABCAMERA_HANDLE hCH, int ImgSetting, int SetType, int Value);

// ABCanonCheckImgSettingWritable return either ABC_RETVAL_OK or ABC_RETVAL_INVALID_PARM
int ABCAPI ABCanonCheckImgSettingWritable (ABCAMERA_HANDLE hCH, int ImgSetting);

// Return value is an error or the number of used entries of pParmArray
int ABCAPI ABCanonGetWritableSettings (ABCAMERA_HANDLE hCH, int32_t *pSettingsArray, int SettingsArraySize);

// Old defines for compatibility. These are the values of the custom functions for the 350D.
// If the camera is a 5D, these values will be automatically mapped to the corresponding custom functions.
#define ABC_CUSTOPT_SETBUTTON_FUNCT      1
#define ABC_CUSTOPT_LONGEXP_NOISE_RED    2
#define ABC_CUSTOPT_FLASH_SYNC           3
#define ABC_CUSTOPT_SHUTTER_AE_LOCK_BTN  4
#define ABC_CUSTOPT_AF_ASSIST_BEAM       5
#define ABC_CUSTOPT_EXP_LEVEL_INCREMENT  6
#define ABC_CUSTOPT_MIRROR_LOCKUP        7
#define ABC_CUSTOPT_E_TTL_II             8
#define ABC_CUSTOPT_SHUTTER_CURTAIN      9

// These new values maps all the possible values.
// An error is returned if the Custom Function requested is not defined for
// the current camera
#define ABC_CF_SETBUTTON_FUNCT           0xAB0000
#define ABC_CF_LONGEXP_NOISE_RED         0xAB0001
#define ABC_CF_FLASH_SYNC                0xAB0002
#define ABC_CF_SHUTTER_AE_LOCK_BTN       0xAB0003
#define ABC_CF_AF_ASSIST_BEAM            0xAB0004
#define ABC_CF_EXP_LEVEL_INCREMENT       0xAB0005
#define ABC_CF_FLASH_FIRING              0xAB0006
#define ABC_CF_ISO_EXPANSION             0xAB0007
#define ABC_CF_AEB_SEQUENCE              0xAB0008
#define ABC_CF_SUPERIMPOSED_DISPLAY      0xAB0009
#define ABC_CF_MENU_BUTTON_DISPLAY       0xAB000A
#define ABC_CF_MIRROR_LOCKUP             0xAB000B
#define ABC_CF_AF_POINT_SELECTION        0xAB000C
#define ABC_CF_E_TTL_II                  0xAB000D
#define ABC_CF_SHUTTER_CURTAIN           0xAB000E
#define ABC_CF_SAFETY_SHIFT              0xAB000F
#define ABC_CF_AF_POINT_ACTIVATION       0xAB0010
#define ABC_CF_RETURN_TO_SHOOT           0xAB0011
#define ABC_CF_LENS_AF_STOP_BUTTON       0xAB0012
#define ABC_CF_ADD_ORIGINAL_DATA         0xAB0013
// Obsolete int ABCAPI ABCanonGetCustomOption (ABCAMERA_HANDLE hCH, int CustomOption);
int ABCAPI ABCanonGetCustomFunction (ABCAMERA_HANDLE hCH, int CustomFunctionNum);

// pList[] == NULL:
// returns the list length in items. ListSize is ignored.
// pList[] != NULL:
// Compiles in pList[] the list of valid values for image setting ImgSetting and returns
// the list length in items, or ABC_RETVAL_BUFFER_TOO_SMALL if ListSize is too small.
//
// Note that if you call first ABCanonGetList to query for the size of the list (using pList[] == NULL) and
// then call ABCanonGetList again with a buffer sized with the value returned by the previous call, you should
// anyway check the returned value because the list size may have changed between the two calls (and so you can
// get fewer items if the list shrunk, or ABC_RETVAL_BUFFER_TOO_SMALL if the list grew).
// Error returned are management errors only (ABC_RETVAL_SEM_ERROR, ABC_RETVAL_INVALID_HANDLE, ABC_RETVAL_INVALID_PARM)
int ABCAPI ABCanonGetList (ABCAMERA_HANDLE hCH, int ImgSetting, uint32_t pList[], size_t ListSize);

// The following function return the serial number, or 0 in case of error
uint32_t ABCAPI ABCanonGetSerialNumber (ABCAMERA_HANDLE hCH);

char * ABCAPI ABCanonGetOwnerName (ABCAMERA_HANDLE hCH);
char * ABCAPI ABCanonGetCameraName (ABCAMERA_HANDLE hCH);
int ABCAPI ABCanonGetFWVersion (ABCAMERA_HANDLE hCH, uint8_t FWRelease[3]);

// Returns ABCTYPE_xxxxx
int ABCAPI ABCanonGetCameraType (ABCAMERA_HANDLE hCH);

// Returns the number of Image types the camera and a pointer to an array of struct sImageTypes describing them
// Or an error.
struct sImageTypes
{
 char     Description[20];
 uint32_t Height;
 uint32_t Width;
 #define  ABC_PT_FINE        2
 #define  ABC_PT_NORMAL      3
 #define  ABC_PT_RAW         4
 uint32_t Type;
};
int ABCAPI ABCanonGetPictureTypes (ABCAMERA_HANDLE hCH, struct sImageTypes **pSIT);

// Returns the number of images for a certain PictureQuality.
// If PictureQuality != -1, the number of images for the given picture quality is returned
// If PictureQuality == -1, the number of images for the CURRENT picture quality is returned
// This function may return ABC_RETVAL_INVALID_PARM if PictureQuality is not a valid value for
//  the camera
int ABCAPI ABCanonGetPQImageCount (ABCAMERA_HANDLE hCH, int32_t PictureQuality);

#define ABC_BATTERY_NORMAL          6
#define ABC_BATTERY_WEAK            4
#define ABC_BATTERY_LOW             3
#define ABC_BATTERY_LB              2
int ABCAPI ABCanonGetBatteryStatus (ABCAMERA_HANDLE hCH);

// If BattLevel == -1 returns the value read from the camera, else converts the code in BattLevel to text
// If BattLevel != -1, hCH may be NULL
char * ABCAPI ABCanonGetBatteryStatusText (ABCAMERA_HANDLE hCH, int BattLevel);

// Camera clock
int ABCAPI ABCanonGetCameraClock (ABCAMERA_HANDLE hCH, time_t *pTime);
int ABCAPI ABCanonSetCameraClock (ABCAMERA_HANDLE hCH, time_t Time);

// Image downloading & C
int ABCAPI ABCanonSetImageCounter (ABCAMERA_HANDLE hCH, uint32_t Counter);
int ABCAPI ABCanonGetImageCounter (ABCAMERA_HANDLE hCH, uint32_t *pCounter);
int ABCAPI ABCanonSetSaveFilenameA (ABCAMERA_HANDLE hCH, char *FileName, int DigitsCount);
int ABCAPI ABCanonSetSaveOverrideA (ABCAMERA_HANDLE hCH, char *FileName);
#ifdef WINVER
 int ABCAPI ABCanonSetSaveFilenameW (ABCAMERA_HANDLE hCH, LPWSTR FileName, int DigitsCount);
 int ABCAPI ABCanonSetSaveOverrideW (ABCAMERA_HANDLE hCH, LPWSTR FileName);
#endif
 // Image download parameters
#define ABC_IMGDL_DISCARD              0
#define ABC_IMGDL_SAVE_TO_FILE         1
#define ABC_IMGDL_SAVE_TO_MEMORY       2
#define ABC_IMGDL_OVERRIDE_OR_DISCARD  3         // 2.02
int ABCAPI ABCanonSetDownloadMode (ABCAMERA_HANDLE hCH, int DownloadMode);

#define ABC_SAVETGT_PC     2
#define ABC_SAVETGT_CF     8
#define ABC_SAVETGT_BOTH   10
int ABCAPI ABCanonSetSaveTarget (ABCAMERA_HANDLE hCH, int SaveTarget);

// Shoot an image
#define ABC_SHOOTINGDENIED_ALLOWED               0
#define ABC_SHOOTINGDENIED_MIRROR_LOCK           1
#define ABC_SHOOTINGDENIED_SELF_TIMER            2
#define ABC_SHOOTINGDENIED_BULB                  3
int ABCAPI ABCanonIsShootingAllowed (ABCAMERA_HANDLE hCH, int *Reason);
int ABCAPI ABCanonShoot (ABCAMERA_HANDLE hCH);

// Options
#define ABC_OPTIONS_BULB_IN_LIST         0x00000001   // 350D Only
#define ABC_OPTIONS_HANDLE_THREAD_BUSY   0x00000002
#define ABC_OPTIONS_STANDARD_BULB        0x00000004   // 5D Only
#define ABC_OPTIONS_ALLOW_OVERWRITE      0x00000008   // 2.02
#define ABC_OPTIONS_PERMANENT_OVERRIDE   0x00000010   // 2.02
#define ABC_OPTIONS_BUSYPOLL             0x00000020   // 2.03
#define ABC_OPTIONS_PCT2                 0x00000040   // 2.03
int ABCAPI ABCanonSetOptions (ABCAMERA_HANDLE hCH, uint32_t NewOptions);
int ABCAPI ABCanonGetOptions (ABCAMERA_HANDLE hCH, uint32_t *pOptions);

// CF
int ABCAPI ABCanonGetRemainingShots (ABCAMERA_HANDLE hCH);
int ABCAPI ABCanonGetCFStatus (ABCAMERA_HANDLE hCH, uint32_t *TotalKB, uint32_t *FreeKB);

// -----------
// Debug functions

typedef void _stdcall (*ABC_DEBUG_STDOUT_CB)(ABCAMERA_HANDLE Handle, void *Msg, uint32_t MsgSize);

#define ABC_STDOUT_LF_ONLY     0
#define ABC_STDOUT_CR_LF       1
#define ABC_STDOUT_ANSI        0
#define ABC_STDOUT_WIDE        2

int ABCAPI ABCDebugSetStdout (int Flags, ABC_DEBUG_STDOUT_CB StdOutFn);
int ABCAPI ABCDebugMemoryUsed (int Reserved);
void ABCAPI ABCDebugDumpHandle (ABCAMERA_HANDLE hCH);

#endif // ABCANON_INCLUDE
