#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>
#include <malloc.h>


#include <windows.h>

#include "ABCanon.h"

#define CAMERA_NUM   4
#define DEFAULT_SAVE_NAME "D:\\1\\ABUTestImg\\IMG_"

// RAW notification are not displayed because they are normally not required.
// Uncomment the following lines if you want to display them.
//#define SHOW_RAW_NOTIFICATIONS


void PrintHelp (void)
{

 puts ("mem             - Show the memory used by the library");
 puts ("scan            - Search for supported cameras");
 puts ("open1           - Connect to first available camera");
 puts ("openidx #       - Connect to camera #");
 puts ("opensn #sn      - Connect to camera with #sn serial number");
 puts ("chdump          - Dump internal camera data");
 puts ("close           - Disconnect from camera");
 puts ("getparm #       - Get image setting value. See ABC_IMGSETTING_xxxx");
 puts ("setparm #       - Change image setting.");
 puts ("getsn           - Get camera serial number");
 puts ("getowner        - Get camera owner");
 puts ("getbatt         - Get battery level");
 puts ("picturetypes    - Show the list of picture types");
 puts ("prmlist #       - Get the list of valid values for setting #");
 puts ("overridefn      - Set override filename");
 puts ("setsavefn       - Set save filename");
 puts ("setimgcounter   - Set image counter for save filename");
 puts ("getimgcounter   - Get current image counter value");
 puts ("downloadmode    - Set download mode to Discard(0), File(1), memory(2),");
 puts ("                  override or discard(3)\n");
 puts ("savetgt         - Save to CF, PC or BOTH");
 puts ("shoot           - Take a picture");
 puts ("shootok         - Check if shooting is allowed");
 puts ("setupshot       - Shortcut to set all the save options at the same time");
 puts ("grevent         - Get an event from the event queue");
 puts ("gettime         - Get camera date/time");
 puts ("settime         - Set camera date/time");
 puts ("setopt          - Get library options (see ABC_OPTIONS_xxx)");
 puts ("getopt          - Set library options (see ABC_OPTIONS_xxx)");
 puts ("custfun #       - Get Custom Function #");
 puts ("cf              - Get CF status/presence in the camera");
 puts ("writable (#)    - Check if image setting # is writable, or return\n"
       "                  the writabel status of all image settings if # is omitted");
 puts ("numofimg (#)    - Numer of files to download for # Picture quality\n");
}

void TranslateEvent (struct sABCEvent *Event)
{

 switch (Event->EventID)
  {
   case ABC_EVENT_NONE:
     printf ("NONE\n");
     break;
   case ABC_EVENT_BATTERY_STATUS_CHANGED:
     printf ("BATTERY_STATUS_CHANGED\n");
     break;
   case ABC_EVENT_SETTINGS_CHANGED:
     printf ("PARAMS_CHANGED\n");
     break;
   case ABC_EVENT_JPG_READY:
     printf ("ABC_EVENT_JPG_READY, #%u\n", Event->EventSize);
     break;
   case ABC_EVENT_CR2_READY:
     printf ("ABC_EVENT_CR2_READY, #%u\n", Event->EventSize);
     break;
   case ABC_EVENT_DOWNLOADING:
     printf ("Img Downloading, %d%%\n", Event->EventSize);
     break;
   case ABC_EVENT_JPG_SAVED:
     printf ("JPG Saved to '%ws', %u bytes\n", Event->EventPtr, Event->EventSize);
     break;
   case ABC_EVENT_CR2_SAVED:
     printf ("CR2 Saved to '%ws', %u bytes\n", Event->EventPtr, Event->EventSize);
     break;
   case ABC_EVENT_JPG_SAVE_FAILED:
     printf ("JPG Image save (%u bytes) failed\n", Event->EventSize);
     break;
   case ABC_EVENT_CR2_SAVE_FAILED:
     printf ("CR2 Image save (%u bytes) failed\n", Event->EventSize);
     break;
   case ABC_EVENT_JPG_IMAGE:
     printf ("JPG Image captured (%u bytes)\n", Event->EventSize);
     break;
   case ABC_EVENT_CR2_IMAGE:
     printf ("CR2 Image captured (%u bytes)\n", Event->EventSize);
     break;
   case ABC_EVENT_AF_FAILED:
     printf ("AF Failed (0x%08X)\n", Event->EventSize);
     break;
   case ABC_EVENT_CAMERA_DISCONNECTED:
     printf ("Camera was disconnected\n");
     break;
   case ABC_EVENT_JPG_DISCARDED:
     printf ("JPG Image discarded\n");
     break;
   case ABC_EVENT_CR2_DISCARDED:
     printf ("CR2 Image discarded\n");
     break;
   default:
      printf ("Unknown event %d\n", Event->EventID);
     break;
  }
}

char *TranslateError (int Error)
{

 switch (Error)
  {
   case ABC_RETVAL_OK:
     return "OK";
   case ABC_RETVAL_ERROR:
     return "ERROR";
   case ABC_RETVAL_BUSY:
     return "CAMERA BUSY";
   case ABC_RETVAL_BUFFER_TOO_SMALL:
     return "BUFFER TOO SMALL";
   case ABC_RETVAL_INVALID_HANDLE:
     return "INVALID HANDLE";
   case ABC_RETVAL_INVALID_PARM:
     return "INVALID PARM";
   case ABC_RETVAL_OUT_OF_MEMORY:
     return "OUT OF MEMORY";
   case ABC_RETVAL_NOT_FOUND:
     return "NOT FOUND";
   case ABC_RETVAL_SEM_ERROR:
     return "SEMAPHORE ERROR";
   case ABC_RETVAL_INVALID_PATH:
     return "INVALID PATH";
   case ABC_RETVAL_TIMEOUT:
     return "TIMEOUT";
   case ABC_RETVAL_DENIED:
     return "DENIED";
   default:
     return "UNKNOWN ERROR";
  }
}

#ifdef SHOW_RAW_NOTIFICATIONS
void _stdcall MyRAWNotificationFunction (ABCAMERA_HANDLE Handle, uint16_t EventID, uint16_t Severity, uint8_t *pData, int DataSize)
{

 // RAW notification are not displayed because they are normally not required.
 // Uncomment the following lines if you want to display them.
 printf ("ABCRAWNotificationFunction called (SN: %u, EventID 0x%04X, Severity 0x%04X, %d bytes\n",
                                           ABCanonGetSerialNumber (Handle), EventID, Severity, DataSize);
}
#endif

int _stdcall MyNotificationFunction (ABCAMERA_HANDLE Handle, struct sABCEvent *pEvent)
{
 struct sDownloadPct *sDP;

 printf ("ABCNotificationFunction called (SN: %u, EventID %d, Ptr 0x%08x, size %d)\n",
                                           ABCanonGetSerialNumber (Handle), pEvent->EventID, pEvent->EventPtr, pEvent->EventSize);

 switch (pEvent->EventID)
  {
   case ABC_EVENT_JPG_SAVED:
   case ABC_EVENT_CR2_SAVED:
     printf ("+ Image saved: %u bytes, %S\n", pEvent->EventSize, (wchar_t *) pEvent->EventPtr);
     break;

   case ABC_EVENT_DOWNLOADING:
     if (pEvent->EventPtr)
      {
       sDP = (struct sDownloadPct *) pEvent->EventPtr;
       printf ("+ Download image %d of %d, %d%% (%d%%)\n", sDP->CurrentImage + 1, sDP->TotalImages,
              pEvent->EventSize, sDP->FullPct);
      }
     else
       printf ("+ Download image, %d%%\n", pEvent->EventSize);
     break;

   case ABC_EVENT_BUSY:
     if (pEvent->EventSize)
      printf ("+ Camera is BUSY\n");
     else
      printf ("+ Camera is no longer busy\n");
     break;

   case ABC_EVENT_ALL_DONE:
     printf ("+ Shoot/Download operation completed!\n");
     break;

   case ABC_EVENT_SHOT_COMPLETED:
     printf ("+ Exposure completed.\n");
     break;
  }

 // To show the Event Queue, this custom notification function returns 0 to all events, so they will be stored in the queue.
 // Obviously, we have to return 1 to the ABC_EVENT_xxx_READY events so the image will not be discarded.
 // If you don't want to use the queue you can either not use a custom notification function or do a function that always
 // returns 1
 switch (pEvent->EventID)
  {
   case ABC_EVENT_JPG_READY:
   case ABC_EVENT_CR2_READY:
     return 1;
   default:
     return 0;
  }
}

void myDump (void *vPtr, int Size)
{
 uint8_t *Ptr = (uint8_t *) vPtr;
 int i, j;

 for (i = 0; i < Size; i += 16)
  {
   printf ("0x%04X:", i);
   int l = i + 16;
   if (l > Size)
    l = Size;
   for (j = i; j < l; j++)
    {
     if ((j & 15) == 8)
      printf (" -");
     printf (" %02X", *(Ptr++));
    }
   printf ("\n");
  }
}

void _stdcall myDebugPrintf (ABCAMERA_HANDLE Handle, void *Msg, uint32_t MsgSize)
{

 printf ("%s", Msg);
}

// ----------------------------------------------------------

char buffer[256];

void Input (void)
{
 char *c;

 memset (buffer, 0, sizeof (buffer));
 fgets (buffer, sizeof (buffer) - 1, stdin);
 if ((c = strchr (buffer, '\n')) != NULL)
  *c = '\0';
}

char *GetParamDescription (int i)
{
 char *Txt;

 switch (i)
  {
   case ABC_IMGSETTING_MODE_DIAL:
     Txt = "MODE_DIAL";
     break;
   case ABC_IMGSETTING_SHUTTER_SPEED:
     Txt = "SHUTTER_SPEED";
     break;
   case ABC_IMGSETTING_APERTURE:
     Txt = "APERTURE";
     break;
   case ABC_IMGSETTING_ISO:
     Txt = "ISO";
     break;
   case ABC_IMGSETTING_EXP_COMPENSATION:
     Txt = "EXP_COMPENSATION";
     break;
   case ABC_IMGSETTING_METERING:
     Txt = "METERING";
     break;
   case ABC_IMGSETTING_AUTOFOCUS:
     Txt = "AUTOFOCUS";
     break;
   case ABC_IMGSETTING_WHITE_BALANCE:
     Txt = "WHITE_BALANCE";
     break;
   case ABC_IMGSETTING_PICTURE_QUALITY:
     Txt = "PICTURE_QUALITY";
     break;
   case ABC_IMGSETTING_FLASH_EXP_COMP:
     Txt = "FLASH_EXP_COMP";
     break;
   case ABC_IMGSETTING_AUTO_POWER_OFF:
     Txt = "AUTO_POWER_OFF";
     break;
   case ABC_IMGSETTING_DRIVE_MODE:
     Txt = "DRIVE_MODE";
     break;
   case ABC_IMGSETTING_AEB:
     Txt = "AEB";
     break;
   case ABC_IMGSETTING_COLOR_TEMPERATURE:
     Txt = "COLOR_TEMPERATURE";
     break;
   default:
     Txt = "???";
     break;
  }

 return Txt;
}

// ----------------------------------------------------------------------------------------

void main (void)
{
 char *Cmd, *Prm;
 ABCAMERA_HANDLE CH[CAMERA_NUM];
 int Idx = 0;

 memset (CH, 0, sizeof (CH));

 uint32_t VersionH, VersionL, Build;
 ABCanonGetVersion (&VersionH, &VersionL, &Build);
 printf ("\nDLL Version: %d.%02d.%d\n\n", VersionH, VersionL, Build);

 ABCDebugSetStdout (ABC_STDOUT_LF_ONLY | ABC_STDOUT_ANSI, myDebugPrintf);

 for (;;)
  {
   printf ("\n>");
   Input ();

   Cmd = strtok (buffer, " \x9");
   Prm = strtok (NULL, "");
   if (Cmd == NULL)
    continue;

   if ((strcmpi (Cmd, "help") == 0) || (strcmpi (Cmd, "?") == 0))
    {
     PrintHelp ();
     continue;
    }

   if (strcmpi (Cmd, "mem") == 0)
    {
     int Prmi = 0;
     if (Prm)
       Prmi = atoi (Prm);
     int rv = ABCDebugMemoryUsed (Prmi);
     printf ("Memory Used %d\n", rv);
     continue;
    }

   if (strcmpi (Cmd, "scan") == 0)
    {
     struct sCameraData *pCD;
     int Num = ABCanonBuildCameraList (&pCD);
     printf ("%d cameras found\n", Num);
     for (int i = 0; i < Num; i++)
      {
       struct sCameraData *CD = (pCD + i);
       printf ("Camera %d: ", i);
       switch (CD->Status)
        {
         case ABC_STS_OK:
           printf ("Type %d, FW: %d.%d.%d, SN: %9u, Name: %s\n", CD->CameraType, CD->FWRelease[0], CD->FWRelease[1],
                    CD->FWRelease[2], CD->SerialNumber, CD->CameraName);
           break;
         case ABC_STS_INUSE:
           printf ("In USE.\n");
           break;
         default:
           printf ("Internal error %d\n", CD->Status);
           break;
        }
      }
     ABCanonReleaseCameraList (pCD);
     continue;
    }

   if (strcmpi (Cmd, "open1") == 0)
    {
     int rv;
     printf ("Opening first camera (Index %d)...\n", Idx);
     rv = ABCanonConnectToCamera (CONNECT_FIRST_AVAILABLE, ABCTYPE_ALL, 0, &CH[Idx]);
     printf ("CameraHandle[%d]: 0x%08X (rv=%d)\n", Idx, CH[Idx], rv);
       if (rv == ABC_RETVAL_OK)
      {
       ABCanonRegisterNotificationCB (CH[Idx], MyNotificationFunction);
       #ifdef SHOW_RAW_NOTIFICATIONS
        ABCanonRegisterRawNotificationCB (CH[Idx], MyRAWNotificationFunction);
       #endif
      }
     continue;
    }

   if (strcmpi (Cmd, "openidx") == 0)
    {
     if (Prm)
      {
       int rv;
       int CIdx = strtol (Prm, NULL, 0);
       printf ("Opening camera #%d (Index %d)...\n", CIdx, Idx);
       rv = ABCanonConnectToCamera (CONNECT_BY_INDEX, ABCTYPE_ALL, CIdx, &CH[Idx]);
       printf ("CameraHandle[%d]: 0x%08X (rv=%d)\n", Idx, CH[Idx], rv);
       if (rv == ABC_RETVAL_OK)
        {
         ABCanonRegisterNotificationCB (CH[Idx], MyNotificationFunction);
         #ifdef SHOW_RAW_NOTIFICATIONS
          ABCanonRegisterRawNotificationCB (CH[Idx], MyRAWNotificationFunction);
         #endif
        }
      }
     continue;
    }

   if (strcmpi (Cmd, "opensn") == 0)
    {
     if (Prm)
      {
       int rv;
       unsigned long SN = strtoul (Prm, NULL, 0);
       printf ("Opening camera with SN %u (Index %d)...\n", SN, Idx);
       rv = ABCanonConnectToCamera (CONNECT_BY_SERIAL_NUMBER, ABCTYPE_ALL, SN, &CH[Idx]);
       printf ("CameraHandle[%d]: 0x%08X (rv=%d)\n", Idx, CH[Idx], rv);
       if (rv == ABC_RETVAL_OK)
        {
         ABCanonRegisterNotificationCB (CH[Idx], MyNotificationFunction);
         #ifdef SHOW_RAW_NOTIFICATIONS
          ABCanonRegisterRawNotificationCB (CH[Idx], MyRAWNotificationFunction);
         #endif
        }
      }
     continue;
    }

   if (strcmpi (Cmd, "select") == 0)
    {
     if (Prm)
      {
       int CIdx = strtol (Prm, NULL, 0);
       if ((CIdx >= 0) && (CIdx < CAMERA_NUM))
        {
         Idx = CIdx;
         printf ("Camera %d selected\n", Idx);
        }
      }
     continue;
    }

   if (strcmpi (Cmd, "chdump") == 0)
    {
     ABCDebugDumpHandle (CH[Idx]);
     continue;
    }

   if (strcmpi (Cmd, "close") == 0)
    {
     printf ("Closing...\n");
     int rv = ABCanonDisconnectFromCamera (CH[Idx]);
     if (ABC_ERROR (rv))
      printf ("ABCanonDisconnectFromCamera returned error %d (%s)\n", rv, TranslateError (rv));
     CH[Idx] = NULL;
     continue;
    }

   // --------------- get/set generic parameter --------------------
   if (strcmpi (Cmd, "getparm") == 0)
    {
     if (Prm)
      {
       int Value = atoi (Prm);
       int Prmi = ABCanonGetImgSetting (CH[Idx], Value);
       if (ABC_ERROR (Prmi))
        printf ("Camera error %d (%s)\n", Prmi, TranslateError (Prmi));
       else
        {
         char *PTxt = ABCanonGetImgSettingText (CH[Idx], Value);
         if (PTxt)
          printf ("Parameter %d is %d (0x%X) - %s\n", Value, Prmi, Prmi, PTxt);
         else
          printf ("Parameter %d is %d (0x%X)\n", Value, Prmi, Prmi);
        }
      }
     else
      {
       int32_t Prms[16];
       int rv = ABCanonGetAllSettings (CH[Idx], Prms, sizeof (Prms) / sizeof (Prms[0]));
       if (ABC_ERROR (rv))
        printf ("ABCanonGetAllSettings returned error %d (%s)\n", rv, TranslateError (rv));
       else
        {
         for (int i = 0; i < rv; i++)
          {
           char *Txt = GetParamDescription (i);
           char *PTxt = ABCanonImgSettingValueToText (i, Prms[i]);
           if (PTxt)
            printf ("Param \"%-30.30s\" (%2d) is %8d (%8X) - %s\n", Txt, i, Prms[i], Prms[i], PTxt);
           else
            printf ("Param \"%-30.30s\" (%2d) is %8d (%8X)\n", Txt, i, Prms[i], Prms[i]);
          }
        }
      }
     continue;
    }
   if (strcmpi (Cmd, "setparm") == 0)
    {
     if (Prm)
      {
       int Prmi = atoi (Prm);
       int PValue = ABCanonGetImgSetting (CH[Idx], Prmi);
       if (!ABC_ERROR (PValue))
        {
         char *PTxt = ABCanonGetImgSettingText (CH[Idx], Prmi);
         printf ("Parameter %d is now %d (%s).", Prmi, PValue, PTxt);
        }
       else
        printf ("Cannot get current value for parameter %d (rv=%d).", Prmi, PValue);

       printf(" New Value (<, >, +, -, P, N): ");
       Input ();
       printf ("\n");
       char *c;
       for (c = buffer; *c == ' '; c++);
       int rv = -99;
       switch (*c)
        {
         case '<':
           printf ("Set Parm %d: First\n", Prmi);
           rv = ABCanonSetImgSetting (CH[Idx], Prmi, ABC_SETTYPE_FIRST, 0);
           break;
         case '>':
           printf ("Set Parm %d: Last\n", Prmi);
           rv = ABCanonSetImgSetting (CH[Idx], Prmi, ABC_SETTYPE_LAST, 0);
           break;
         case '+':
         case '-':
           printf ("Set Parm %d: rc +%d\n", Prmi, atoi (c));
           rv = ABCanonSetImgSetting (CH[Idx], Prmi, ABC_SETTYPE_CHANGE, atoi (c));
           break;
         case 'P':
         case 'p':
           printf ("Set Parm %d: Previous\n", Prmi);
           rv = ABCanonSetImgSetting (CH[Idx], Prmi, ABC_SETTYPE_PREV, 0);
           break;
         case 'N':
         case 'n':
           printf ("Set Parm %d: Next\n", Prmi);
           rv = ABCanonSetImgSetting (CH[Idx], Prmi, ABC_SETTYPE_NEXT, 0);
           break;
         default:
           printf ("Set Parm %d: abs %d\n", Prmi, atoi (c));
           rv = ABCanonSetImgSetting (CH[Idx], Prmi, ABC_SETTYPE_SET, atoi (c));
           break;
        }
       printf ("Param set: rv = %d\n", rv);
      }
     else
      printf ("Missing parameter value!!\n");

     continue;
    }
   // --------------- check generic functions --------------------
   if (strcmpi (Cmd, "getsn") == 0)
    {
     printf ("Camera SN: %u\n", ABCanonGetSerialNumber (CH[Idx]));
     continue;
    }
   if (strcmpi (Cmd, "getowner") == 0)
    {
     printf ("Camera owner: %s\n", ABCanonGetOwnerName (CH[Idx]));
     continue;
    }
   if (strcmpi (Cmd, "getbatt") == 0)
    {
     printf ("Battery level %s (%d)\n", ABCanonGetBatteryStatusText (CH[Idx], -1), ABCanonGetBatteryStatus (CH[Idx]));
     continue;
    }
   if (strcmpi (Cmd, "picturetypes") == 0)
    {
     struct sImageTypes *pSIT;
     int PictureTypes = ABCanonGetPictureTypes (CH[Idx], &pSIT);
     printf ("Picture types: %d\n", PictureTypes);
     for (int i = 0; i < PictureTypes; i++)
      printf ("%2d: type %d, %4dx%4d, %s\n", i, pSIT[i].Type, pSIT[i].Width, pSIT[i].Height, pSIT[i].Description);
     continue;
    }

   // --------------- lists --------------------
   if (strcmpi (Cmd, "prmlist") == 0)
    {
     if (Prm)
      {
       int Prmi = atoi (Prm);
       int ListSize = ABCanonGetList (CH[Idx], Prmi, NULL, 0);
       printf ("List size for parameter %d is %d\n", Prmi, ListSize);
       if (!ABC_ERROR (ListSize))
        {
         uint32_t *pu32 = malloc (ListSize * sizeof (uint32_t));
         if (pu32)
          {
           ListSize = ABCanonGetList (CH[Idx], Prmi, pu32, ListSize);
           if (!ABC_ERROR (ListSize))
            {
             printf ("%d items:", ListSize);
             for (int i = 0; i < ListSize; i++)
              printf (" 0x%X", *(pu32 + i));
             printf ("\n");
            }
           free (pu32);
          }
        }
      }
     continue;
    }

   // --------------- Filenames/ImgCounter --------------------
   if (strcmpi (Cmd, "overridefn") == 0)
    {
     if (Prm != NULL)
      {
       int rv = ABCanonSetSaveOverrideA (CH[Idx], Prm);
       printf ("ABCanonSetSaveOverrideA('%s'): rv = %d\n", Prm, rv);
      }
     else
      {
       int rv = ABCanonSetSaveOverrideA (CH[Idx], NULL);
       printf ("ABCanonSetSaveOverrideA(NULL): rv = %d\n", rv);
      }
     continue;
    }
   if (strcmpi (Cmd, "setsavefn") == 0)
    {
     if (Prm == NULL)
      Prm = DEFAULT_SAVE_NAME;

     int rv = ABCanonSetSaveFilenameA (CH[Idx], Prm, 4);
     uint32_t IC;
     ABCanonGetImageCounter (CH[Idx], &IC);
     printf ("ABCanonSetSaveFilenameA: rv = %d, ImgCounter = %d\n", rv, IC);
     continue;
    }
   if (strcmpi (Cmd, "setimgcounter") == 0)
    {
     if (Prm)
      {
       int Prmi = atoi (Prm);
       int rv = ABCanonSetImageCounter (CH[Idx], Prmi);
       uint32_t IC;
       ABCanonGetImageCounter (CH[Idx], &IC);
       printf ("ABCanonSetImageCounter: rv = %d, ImgCounter = %d\n", rv, IC);
      }
     continue;
    }
   if (strcmpi (Cmd, "getimgcounter") == 0)
    {
     uint32_t IC;
     int rv = ABCanonGetImageCounter (CH[Idx], &IC);
     printf ("ABCanonGetImageCounter: rv = %d, ImgCounter = %d\n", rv, IC);
     continue;
    }
   if (strcmpi (Cmd, "downloadmode") == 0)
    {
     if (Prm)
      {
       int Prmi = atoi (Prm);
       int rv = ABCanonSetDownloadMode (CH[Idx], Prmi);
       printf ("ABCanonSetDownloadMode: rv = %d\n", rv);
      }
     continue;
    }

   if (strcmpi (Cmd, "savetgt") == 0)
    {
     if (Prm)
      {
       if (strcmpi (Prm, "CF") == 0)
        {
         int rv = ABCanonSetSaveTarget (CH[Idx], ABC_SAVETGT_CF);
         printf ("ABCanonSetSaveTarget to CF, rv=%d\n", rv);
        }
       else if (strcmpi (Prm, "PC") == 0)
        {
         int rv = ABCanonSetSaveTarget (CH[Idx], ABC_SAVETGT_PC);
         printf ("ABCanonSetSaveTarget to PC, rv=%d\n", rv);
        }
       else if (strcmpi (Prm, "BOTH") == 0)
        {
         int rv = ABCanonSetSaveTarget (CH[Idx], ABC_SAVETGT_PC | ABC_SAVETGT_CF);
         printf ("ABCanonSetSaveTarget to PC and CF, rv=%d\n", rv);
        }
      }
     continue;
    }

   if (strcmpi (Cmd, "shoot") == 0)
    {
     int rv = ABCanonShoot (CH[Idx]);
     printf ("ABCanonShot, rv=%d\n", rv);
     continue;
    }

   if (strcmpi (Cmd, "shootok") == 0)
    {
     int Reason = -1;
     int rv = ABCanonIsShootingAllowed (CH[Idx], &Reason);
     printf ("ABCanonIsShootingAllowed, rv=%d, reason: %d\n", rv, Reason);
     continue;
    }

   if (strcmpi (Cmd, "setupshot") == 0)
    {
     char fn[256];
     sprintf (fn, DEFAULT_SAVE_NAME "%d_", Idx);
     ABCanonSetSaveFilenameA (CH[Idx], fn, 4);
     ABCanonSetDownloadMode (CH[Idx], ABC_IMGDL_SAVE_TO_FILE);
     ABCanonSetSaveTarget (CH[Idx], ABC_SAVETGT_PC);
     printf ("setupshot done\n");
     continue;
    }

 // ------------------------------------------------------------------

   if (strcmpi (Cmd, "grevent") == 0)
    {
     struct sABCEvent Evt;
     int rv = ABCanonGetEvent (CH[Idx], &Evt, 100);
     if (rv == ABC_RETVAL_TIMEOUT)
      printf ("No events...\n");
     else if (rv != ABC_RETVAL_OK)
      printf ("ABCanonGetEvent rv=%d\n", rv);
     else
      {
       printf ("Evt: %d, %u@0x%08X\n", Evt.EventID, Evt.EventSize, Evt.EventPtr);
       if (Evt.EventPtr != NULL)                              // _msize
        {
         int s = _msize (Evt.EventPtr);
         myDump (Evt.EventPtr, (s > 64) ? 64 : s);
        }
       printf (" * ");
       TranslateEvent (&Evt);
       ABCanonReleaseEvent (&Evt);
      }
     continue;
    }

 // ------------------------------------------------------------------

   if (strcmpi (Cmd, "gettime") == 0)
    {
     time_t t;
     int rv = ABCanonGetCameraClock (CH[Idx], &t);
     printf ("ABCanonGetCameraClock: rv=%d, time=%s\n", rv, ctime (&t));
     continue;
    }

   if (strcmpi (Cmd, "settime") == 0)
    {
     time_t t;
     time (&t);
     int rv = ABCanonSetCameraClock (CH[Idx], t);
     printf ("ABCanonSetCameraClock: rv=%d, time=0x%08x -> %s\n", rv, t, ctime (&t));
     continue;
    }

 // ------------------------------------------------------------------

   if (strcmpi (Cmd, "setopt") == 0)
    {
     if (Prm)
      {
       uint32_t Prmi = strtoul (Prm, NULL, 0);
       int rv = ABCanonSetOptions (CH[Idx], Prmi);
       printf ("ABCanonSetOptions (0x%08X): rv=%d\n", Prmi, rv);
      }
     continue;
    }

   if (strcmpi (Cmd, "getopt") == 0)
    {
     uint32_t Prmi = 0;
     int rv = ABCanonGetOptions (CH[Idx], &Prmi);
     printf ("ABCanonGetOptions (): rv=%d, Options=0x%08X\n", rv, Prmi);
     continue;
    }

 // ------------------------------------------------------------------

   if (strcmpi (Cmd, "custfun") == 0)
    {
     if (Prm)
      {
       uint32_t Prmi = strtoul (Prm, NULL, 0);
       int rv = ABCanonGetCustomFunction (CH[Idx], Prmi);
       if (ABC_ERROR (rv))
        printf ("ABCanonGetCustomFunction returned error %d\n", rv);
       else if (Prmi < 10)
        printf ("Custom option %d is %d\n", Prmi, rv);
       else
        printf ("Custom option 0x%06x is %d\n", Prmi, rv);
      }
     continue;
    }

 // ------------------------------------------------------------------

   if (strcmpi (Cmd, "cf") == 0)
    {
     uint32_t Total, Free;
     int rv = ABCanonGetCFStatus (CH[Idx], &Total, &Free);
     printf ("ABCanonGetCFStatus (): rv=%d, Total %u kB, Free %u kB\n", rv, Total, Free);
     continue;
    }

 // ------------------------------------------------------------------

   if (strcmpi (Cmd, "writable") == 0)
    {
     if (Prm)
      {
       uint32_t Prmi = strtoul (Prm, NULL, 0);
       int rv = ABCanonCheckImgSettingWritable (CH[Idx], Prmi);
       printf ("ABCanonCheckImgSettingWritable (%d): rv=%d\n", Prmi, rv);
      }
     else
      {
       int32_t Prms[16];
       int rv = ABCanonGetWritableSettings (CH[Idx], Prms, sizeof (Prms) / sizeof (Prms[0]));
       if (ABC_ERROR (rv))
        printf ("ABCanonGetWritableSettings returned error %d\n", rv);
       else
        {
         for (int i = 0; i < rv; i++)
          {
           char *Txt = GetParamDescription (i);
           printf ("Param \"%-30.30s\" (%2d)  Writable %d\n", Txt, i, Prms[i]);
          }
        }
      }
     continue;
    }

 // ------------------------------------------------------------------

   if (strcmpi (Cmd, "numofimg") == 0)
    {
     if (Prm)
      {
       uint32_t Prmi = strtoul (Prm, NULL, 0);
       int rv = ABCanonGetPQImageCount (CH[Idx], Prmi);
       printf ("ABCanonGetPQImageCount (%d): rv=%d\n", Prmi, rv);
      }
     else
      {
       int rv = ABCanonGetPQImageCount (CH[Idx], -1);
       printf ("ABCanonGetPQImageCount (-1): rv=%d\n", rv);
      }
     continue;
    }



  }

}
