﻿Imports System.Runtime.InteropServices
Imports System.Threading
Public Class ABCanon7
 Implements IDisposable

 ' https://docs.microsoft.com/it-it/dotnet/visual-basic/programming-guide/language-features/events/how-to-declare-custom-events-to-avoid-blocking
 ' https://docs.microsoft.com/en-us/dotnet/api/system.collections.arraylist.synchronized?view=netframework-4.5#System_Collections_ArrayList_Synchronized_System_Collections_IList_
 ' https://blog.stephencleary.com/2013/10/taskrun-etiquette-and-proper-usage.html

#If DEBUG Then
 Private Const DLL As String = "..\..\ABCanon7.dll"
#Else
   Private Const DLL As String = "ABCanon7.dll"
#End If

 Public Enum eEventID As UInteger
  ABC_EVENT_NONE = 0
  ABC_EVENT_BATTERY_STATUS_CHANGED = 1
  ABC_EVENT_SETTINGS_CHANGED = 2
  ABC_EVENT_SHOT_COMPLETED = 3
  ABC_EVENT_JPG_READY = 4
  ABC_EVENT_CR2_READY = 5
  ABC_EVENT_DOWNLOADING = 6
  ABC_EVENT_JPG_SAVED = 7
  ABC_EVENT_CR2_SAVED = 8
  ABC_EVENT_JPG_SAVE_FAILED = 9
  ABC_EVENT_CR2_SAVE_FAILED = 10
  ABC_EVENT_JPG_IMAGE = 11
  ABC_EVENT_CR2_IMAGE = 12
  ABC_EVENT_AF_FAILED = 13
  ABC_EVENT_CAMERA_DISCONNECTED = 14
  ABC_EVENT_JPG_DISCARDED = 15
  ABC_EVENT_CR2_DISCARDED = 16
  ABC_EVENT_BUSY = 17
  ABC_EVENT_DOWNLOAD_COMPLETED = 18
 End Enum
 <StructLayout(LayoutKind.Sequential)> _
 Private Structure sEventInternal
  Dim EventID As eEventID
  Dim EventSize As UInteger
  Dim EventPtr As IntPtr
 End Structure
 Public Structure sEvent
  Dim EventID As eEventID
  Dim EventSize As UInteger
  Dim EventPtr As IntPtr
  Dim Handled As Boolean
 End Structure

 Public Event CameraEvent(Camera As ABCanon7, ByRef CameraEvent As sEvent)
 Public Event [Error](Camera As ABCanon7)

 Private CameraHandle As IntPtr = 0
 Private Disposed As Boolean
 Private gShotParms() As UInteger
 Private gWritableParms() As Boolean

 Private gTCSs As New List(Of TaskCompletionSource(Of Object))
 Private gShootTCS As TaskCompletionSource(Of Object)

 Public Enum eErrors As Integer
  ABC_RETVAL_OK = 0
  ABC_RETVAL_ERROR = -1
  ABC_RETVAL_BUSY = -2
  ABC_RETVAL_BUFFER_TOO_SMALL = -3
  ABC_RETVAL_INVALID_HANDLE = -4
  ABC_RETVAL_INVALID_PARM = -5
  ABC_RETVAL_OUT_OF_MEMORY = -6
  ABC_RETVAL_NOT_FOUND = -7
  ABC_RETVAL_SEM_ERROR = -8
  ABC_RETVAL_INVALID_PATH = -9
  ABC_RETVAL_TIMEOUT = -10
  ABC_RETVAL_DENIED = -11
  ABC_RETVAL_NOT_CONNECTED = -12
  ABC_RETVAL_THREAD_BUSY = -13
  ABC_RETVAL_CANCELLED = -99
 End Enum
 Private gError As eErrors
 Public Property CameraError As eErrors

  Get
   Return gError
  End Get
  Private Set(value As eErrors)
   If value <> gError Then
    gError = value
    RaiseEvent Error(Me)
   End If
  End Set

 End Property
 Public Shared Function IsError(ByVal a As Integer) As Boolean

  If a < 0 Then
   Return True
  Else
   Return False
  End If

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Sub ABCanonGetVersion(ByRef VersionH As UInt32, ByRef VersionL As UInt32, ByRef Build As UInt32)
 End Sub
 Public Shared ReadOnly Property Version As String

  Get
   Dim VersionH As UInt32
   Dim VersionL As UInt32
   Dim Build As UInt32
   ABCanonGetVersion(VersionH, VersionL, Build)
   Return VersionH.ToString & "." & VersionL.ToString("00") & "." & Build
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 Public Enum eCameraStatus As UInt32
  ABC_STS_OK = 0
  ABC_STS_INUSE = 1
  ABC_STS_CREATEFILE = 2
  ABC_STS_WINUSBINIT = 3
  ABC_STS_QUERYDEVINFO = 4
  ABC_STS_EPNUMBER = 5
  ABC_STS_EPTYPE_0 = 6
  ABC_STS_EPTYPE_1 = 7
  ABC_STS_EPTYPE_2 = 8
  ABC_STS_CONTROLRW = 9
  ABC_STS_INITSTATUS = 10
  ABC_STS_BADSIGNATURE = 11
  ABC_STS_UNKNOWNCAMERA = 12
  ABC_STS_CMD12CHECKSFAIL = 13
  ABC_STS_UNKNOWNERROR = 14
  ABC_STS_CMD2512CHECKSFAIL = 15
  ABC_STS_SETPIPEPOLICY = 16
 End Enum
 Public Enum eCameraType As Byte
  ABCTYPE_UNKNOWN = 0
  ABCTYPE_ALL = 0
  ABCTYPE_350D = 1
  ABCTYPE_5D = 2
 End Enum
 <StructLayout(LayoutKind.Sequential)> _
 Public Structure sCameraData
  Dim Idx As UInt32
  Dim Status As eCameraStatus
  Dim SerialNumber As UInt32
  <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=32)> _
  Dim CameraName As String
  Dim VersionH As Byte
  Dim VersionM As Byte
  Dim VersionL As Byte
  Dim CameraType As eCameraType
 End Structure
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonBuildCameraList(ByRef ITPtr As IntPtr) As Integer
 End Function
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Sub ABCanonReleaseCameraList(ByVal ITPtr As IntPtr)
 End Sub
 Public Shared Function GetConnectedCameras() As sCameraData()

  Dim Ptr As IntPtr
  Dim Cameras As Integer = ABCanonBuildCameraList(Ptr)
  ' In case of error (or no cameras) return an empty list
  If IsError(Cameras) Or Cameras = 0 Then
   Return Nothing
  End If

  Dim rv(Cameras - 1) As sCameraData

  For i As Integer = 0 To rv.Length - 1
   rv(i) = Marshal.PtrToStructure(Ptr + Marshal.SizeOf(GetType(sCameraData)) * i, GetType(sCameraData))
  Next

  ABCanonReleaseCameraList(Ptr)

  Return rv

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 Public Enum eConnectMode As UInt32
  FirstAvailable = 0
  ByIndex = 1
  BySerialNumber = 2
 End Enum
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonConnectToCamera(ByVal ConnectMode As eConnectMode, ByVal Filter As UInt32,
                                                 ByVal Param As UInt32, ByRef CameraHandle As IntPtr) As Integer
 End Function
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonRegisterNotificationCB(ByVal CameraHandle As IntPtr, ByVal FuncPtr As ABC_NOTIFICATION_CB) As Integer
 End Function
 Public Sub New(ByVal ConnectMode As eConnectMode, ByVal Param As UInt32)

  Me.New(ConnectMode, eCameraType.ABCTYPE_ALL, Param)

 End Sub
 Public Sub New(ByVal ConnectMode As eConnectMode, ByVal Filter As eCameraType, ByVal Param As UInt32)

  Dim Rv As Integer = ABCanonConnectToCamera(ConnectMode, Filter, Param, CameraHandle)

  Debug.Print("ABCanonConnectToCamera => rv={0}, CameraHandle={1}", Rv, CameraHandle)

  If IsError(Rv) Then
   Dim E As eErrors = Rv
   Throw New Exception(Rv.ToString)
  End If

  ' Register notification function
  ABCanonRegisterNotificationCB(CameraHandle, MyNotificationFunc)

  ' Since this code is meant to be used with library version 2.03 or greater, add some options we use
  ABCanonSetOptions(CameraHandle, eOptions.BUSYPOLL Or eOptions.PCT2)

  CameraError = eErrors.ABC_RETVAL_OK

 End Sub
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Sub ABCanonDisconnectFromCamera(ByVal CameraHandle As IntPtr)
 End Sub
 Public Overloads Sub Dispose() Implements IDisposable.Dispose

  Debug.Print("Closing camera..:")
  If CameraHandle <> 0 Then
   ABCanonRegisterNotificationCB(CameraHandle, Nothing)
   ABCanonDisconnectFromCamera(CameraHandle)
   CameraHandle = 0
  End If
  GC.SuppressFinalize(Me)

 End Sub
 Protected Overrides Sub Finalize()

  Dispose()

 End Sub
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCDebugMemoryUsed(ByVal Reserved As Integer) As Integer
 End Function
 Public Shared Function DbgUsedMemory() As Integer

  Return ABCDebugMemoryUsed(0)

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetSerialNumber(ByVal CameraHandle As IntPtr) As UInt32
 End Function
 Public ReadOnly Property SerialNumber As UInt32

  Get
   Return ABCanonGetSerialNumber(CameraHandle)
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetCameraType(ByVal CameraHandle As IntPtr) As UInt32
 End Function
 Public ReadOnly Property CameraType As eCameraType

  Get
   Return ABCanonGetCameraType(CameraHandle)
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetCameraName(ByVal CameraHandle As IntPtr) As IntPtr
 End Function
 Public ReadOnly Property CameraName As String

  Get
   Dim StrPtr As IntPtr = ABCanonGetCameraName(CameraHandle)
   If StrPtr = Nothing Then
    Return Nothing
   End If
   Return Marshal.PtrToStringAnsi(StrPtr)
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetOwnerName(ByVal CameraHandle As IntPtr) As IntPtr
 End Function
 Public ReadOnly Property OwnerName As String

  Get
   Dim StrPtr As IntPtr = ABCanonGetOwnerName(CameraHandle)
   If StrPtr = Nothing Then
    Return Nothing
   End If
   Return Marshal.PtrToStringAnsi(StrPtr)
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetFWVersion(ByVal CameraHandle As IntPtr, ByVal FWVersion() As Byte) As Integer
 End Function
 Public ReadOnly Property FWVersion As Byte()

  Get
   Dim b(3) As Byte
   ABCanonGetFWVersion(CameraHandle, b)
   Return b
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 Public Enum eImgSetting As Integer
  ABC_IMGSETTING_MODE_DIAL = 0
  ABC_IMGSETTING_SHUTTER_SPEED = 1
  ABC_IMGSETTING_APERTURE = 2
  ABC_IMGSETTING_ISO = 3
  ABC_IMGSETTING_EXP_COMPENSATION = 4
  ABC_IMGSETTING_METERING = 5
  ABC_IMGSETTING_AUTOFOCUS = 6
  ABC_IMGSETTING_WHITE_BALANCE = 7
  ABC_IMGSETTING_PICTURE_QUALITY = 8
  ABC_IMGSETTING_FLASH_EXP_COMP = 9
  ABC_IMGSETTING_AUTO_POWER_OFF = 10
  ABC_IMGSETTING_DRIVE_MODE = 11
  ABC_IMGSETTING_AEB = 12
  ABC_IMGSETTING_COLOR_TEMPERATURE = 13    ' 5D only
 End Enum
 Public Enum eSPModeDial As Integer
  ' Basic zone
  ABC_DIALMODE_AUTO = &H0              ' Full Auto
  ' Creative zone
  ABC_DIALMODE_PROGRAM = &H101         ' Program AE
  ABC_DIALMODE_Tv = &H102              ' Shutter priority AE
  ABC_DIALMODE_Av = &H103              ' Aperture priority AE
  ABC_DIALMODE_MANUAL = &H104          ' Manual exposure
  ABC_DIALMODE_A_DEP = &H105           ' Automatic Depth-of-field AE
  ABC_DIALMODE_BULB = &H107            ' BULB
  ' Image zone
  ABC_DIALMODE_PORTRAIT = &H800        ' Portrait
  ABC_DIALMODE_LANDSCAPE = &H200       ' Landscape
  ABC_DIALMODE_CLOSEUP = &HA00         ' Close-up
  ABC_DIALMODE_SPORTS = &H900          ' Sports
  ABC_DIALMODE_NIGHT_PORTRAIT = &H500  ' Night portrait
  ABC_DIALMODE_NO_FLASH = &HF00        ' No flash
 End Enum
 Public Enum eSPWhiteBalance As Integer
  ABC_WB_AUTO = &H0
  ABC_WB_DAY_LIGHT = &H1
  ABC_WB_CLOUDY = &H2
  ABC_WB_TUNGSTEN = &H3
  ABC_WB_FLUORESCENT = &H4
  ABC_WB_FLASH = &H5
  ABC_WB_CUSTOM = &H6
  ABC_WB_SHADE = &H8
  ABC_WB_K = &H9                       ' 5D Only
 End Enum
 Public Enum eSPMetering As Integer
  ABC_METERING_SPOT = 1                      ' 5D Only
  ABC_METERING_EVALUATIVE = 3
  ABC_METERING_PARTIAL = 4
  ABC_METERING_CENTER_WEIGHTED = 5
 End Enum
 Public Enum eSPPictureQuality As Integer
  ABC_PICTUREQUALITY_L_FINE = &H10320             ' L Fine
  ABC_PICTUREQUALITY_L_NORMAL = &H10220           ' L Normal
  ABC_PICTUREQUALITY_M_FINE = &H1010320           ' M Fine
  ABC_PICTUREQUALITY_M_NORMAL = &H1010220         ' M Normal
  ABC_PICTUREQUALITY_S_FINE = &H2010320           ' S Fine
  ABC_PICTUREQUALITY_S_NORMAL = &H2010220         ' S Normal
  ABC_PICTUREQUALITY_RAW_L_FINE = &H123420        ' RAW+L
  ABC_PICTUREQUALITY_RAW_L_NORMAL = &H122420
  ABC_PICTUREQUALITY_RAW_M_FINE = &H10123420
  ABC_PICTUREQUALITY_RAW_M_NORMAL = &H10122420
  ABC_PICTUREQUALITY_RAW_S_FINE = &H20123420
  ABC_PICTUREQUALITY_RAW_S_NORMAL = &H20122420
  ABC_PICTUREQUALITY_RAW = &H20420                ' RAW
 End Enum
 Public Enum eSPDriveMode As Integer
  ABC_DRIVEMODE_SINGLE = &H0
  ABC_DRIVEMODE_CONTINUOUS = &H1
  ABC_DRIVEMODE_SELF_TIMER = &H100
 End Enum
 Public Enum eSPISO As Integer
  ABC_ISO_100 = &H48
  ABC_ISO_200 = &H50
  ABC_ISO_400 = &H58
  ABC_ISO_800 = &H60
  ABC_ISO_1600 = &H68
 End Enum
 Public Enum eSPAperture3 As Integer
  ABC_APERTURE3_1_0 = &H8
  ABC_APERTURE3_1_1 = &HB
  ABC_APERTURE3_1_2 = &HD
  ABC_APERTURE3_1_4 = &H10
  ABC_APERTURE3_1_6 = &H13
  ABC_APERTURE3_1_8 = &H15
  ABC_APERTURE3_2_0 = &H18
  ABC_APERTURE3_2_2 = &H1B
  ABC_APERTURE3_2_5 = &H1D
  ABC_APERTURE3_2_8 = &H20
  ABC_APERTURE3_3_2 = &H23
  ABC_APERTURE3_3_5 = &H25
  ABC_APERTURE3_4_0 = &H28
  ABC_APERTURE3_4_5 = &H2B
  ABC_APERTURE3_5_0 = &H2D
  ABC_APERTURE3_5_6 = &H30
  ABC_APERTURE3_6_3 = &H33
  ABC_APERTURE3_7_1 = &H35
  ABC_APERTURE3_8_0 = &H38
  ABC_APERTURE3_9_0 = &H3B
  ABC_APERTURE3_10_0 = &H3D
  ABC_APERTURE3_11_0 = &H40
  ABC_APERTURE3_13_0 = &H43
  ABC_APERTURE3_14_0 = &H45
  ABC_APERTURE3_16_0 = &H48
  ABC_APERTURE3_18_0 = &H4B
  ABC_APERTURE3_20_0 = &H4D
  ABC_APERTURE3_22_0 = &H50
  ABC_APERTURE3_25_0 = &H53
  ABC_APERTURE3_29_0 = &H55
  ABC_APERTURE3_32_0 = &H58
  ABC_APERTURE3_36_0 = &H5B
  ABC_APERTURE3_40_0 = &H5D
  ABC_APERTURE3_45_0 = &H60
  ABC_APERTURE3_51_0 = &H63
  ABC_APERTURE3_57_0 = &H65
  ABC_APERTURE3_64_0 = &H68
  ABC_APERTURE3_72_0 = &H6B
  ABC_APERTURE3_81_0 = &H6D
  ABC_APERTURE3_91_0 = &H70
 End Enum
 Public Enum eSPShutterSpeed3 As Integer
  ABC_SHUTTERSPEED3_BULB = 4
  ABC_SHUTTERSPEED3_30s = &H10
  ABC_SHUTTERSPEED3_25s = &H13
  ABC_SHUTTERSPEED3_20s = &H15
  ABC_SHUTTERSPEED3_15s = &H18
  ABC_SHUTTERSPEED3_13s = &H1B
  ABC_SHUTTERSPEED3_10s = &H1D
  ABC_SHUTTERSPEED3_8s = &H20
  ABC_SHUTTERSPEED3_6s = &H23
  ABC_SHUTTERSPEED3_5s = &H25
  ABC_SHUTTERSPEED3_4s = &H28
  ABC_SHUTTERSPEED3_3s2 = &H2B
  ABC_SHUTTERSPEED3_2s5 = &H2D
  ABC_SHUTTERSPEED3_2s = &H30
  ABC_SHUTTERSPEED3_1s6 = &H33
  ABC_SHUTTERSPEED3_1s3 = &H35
  ABC_SHUTTERSPEED3_1s = &H38
  ABC_SHUTTERSPEED3_0s8 = &H3B
  ABC_SHUTTERSPEED3_0s6 = &H3D
  ABC_SHUTTERSPEED3_0s5 = &H40
  ABC_SHUTTERSPEED3_0s4 = &H43
  ABC_SHUTTERSPEED3_0s3 = &H45
  ABC_SHUTTERSPEED3_1_4 = &H48
  ABC_SHUTTERSPEED3_1_5 = &H4B
  ABC_SHUTTERSPEED3_1_6 = &H4D
  ABC_SHUTTERSPEED3_1_8 = &H50
  ABC_SHUTTERSPEED3_1_10 = &H53
  ABC_SHUTTERSPEED3_1_13 = &H55
  ABC_SHUTTERSPEED3_1_15 = &H58
  ABC_SHUTTERSPEED3_1_20 = &H5B
  ABC_SHUTTERSPEED3_1_25 = &H5D
  ABC_SHUTTERSPEED3_1_30 = &H60
  ABC_SHUTTERSPEED3_1_40 = &H63
  ABC_SHUTTERSPEED3_1_50 = &H65
  ABC_SHUTTERSPEED3_1_60 = &H68
  ABC_SHUTTERSPEED3_1_80 = &H6B
  ABC_SHUTTERSPEED3_1_100 = &H6D
  ABC_SHUTTERSPEED3_1_125 = &H70
  ABC_SHUTTERSPEED3_1_160 = &H73
  ABC_SHUTTERSPEED3_1_200 = &H75
  ABC_SHUTTERSPEED3_1_250 = &H78
  ABC_SHUTTERSPEED3_1_320 = &H7B
  ABC_SHUTTERSPEED3_1_400 = &H7D
  ABC_SHUTTERSPEED3_1_500 = &H80
  ABC_SHUTTERSPEED3_1_640 = &H83
  ABC_SHUTTERSPEED3_1_800 = &H85
  ABC_SHUTTERSPEED3_1_1000 = &H88
  ABC_SHUTTERSPEED3_1_1250 = &H8B
  ABC_SHUTTERSPEED3_1_1600 = &H8D
  ABC_SHUTTERSPEED3_1_2000 = &H90
  ABC_SHUTTERSPEED3_1_2500 = &H93
  ABC_SHUTTERSPEED3_1_3200 = &H95
  ABC_SHUTTERSPEED3_1_4000 = &H98
  ABC_SHUTTERSPEED3_1_5000 = &H9B
  ABC_SHUTTERSPEED3_1_6400 = &H9D
  ABC_SHUTTERSPEED3_1_8000 = &HA0
  ABC_SHUTTERSPEED3_1_10000 = &HA3
  ABC_SHUTTERSPEED3_1_12800 = &HA5
  ABC_SHUTTERSPEED3_1_16000 = &HA8
 End Enum
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonGetImgSetting(ByVal CameraHandle As IntPtr, ByVal ImgSetting As eImgSetting) As Integer
 End Function
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonSetImgSetting(ByVal CameraHandle As IntPtr, ByVal ImgSetting As UInteger,
                                                   ByVal SetType As eSetType, ByVal Value As Integer) As Integer
 End Function
 Public Property ImgSetting(ByVal ShotParamIdx As eImgSetting) As eErrors

  Get
   Dim rv As eErrors = ABCanonGetImgSetting(CameraHandle, ShotParamIdx)
   ' Returned values ABC_RETVAL_SEM_ERROR, ABC_RETVAL_INVALID_HANDLE, ABC_RETVAL_INVALID_PARM
   ' should not occur so we handle them with an exception
   Select Case rv
    Case eErrors.ABC_RETVAL_SEM_ERROR, eErrors.ABC_RETVAL_INVALID_HANDLE, eErrors.ABC_RETVAL_INVALID_PARM
     Throw New Exception(rv.ToString)
    Case Else
     If IsError(rv) Then
      CameraError = rv
     Else
      CameraError = eErrors.ABC_RETVAL_OK
     End If
   End Select
   Return rv
  End Get
  Set(value As eErrors)
   Dim rv As eErrors = ABCanonSetImgSetting(CameraHandle, ShotParamIdx, eSetType.SetToValue, value)
   If IsError(rv) Then
    CameraError = rv
   Else
    CameraError = eErrors.ABC_RETVAL_OK
   End If
  End Set

 End Property
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonGetAllSettings(ByVal CameraHandle As IntPtr, ByVal SettingsArray() As UInteger, ByVal SettingsArraySize As UInteger) As Integer
 End Function
 Public ReadOnly Property AllImgSettings As UInteger()

  Get
   Dim Length As Integer = [Enum].GetNames(GetType(eImgSetting)).Length
   If IsNothing(gShotParms) OrElse gShotParms.Length <> Length Then
    ReDim gShotParms(Length - 1)
   End If

   Dim rv As eErrors = ABCanonGetAllSettings(CameraHandle, gShotParms, Length)
   ' Returned values ABC_RETVAL_SEM_ERROR, ABC_RETVAL_INVALID_HANDLE, ABC_RETVAL_INVALID_PARM
   ' should not occur so we handle them with an exception
   Select Case rv
    Case eErrors.ABC_RETVAL_SEM_ERROR, eErrors.ABC_RETVAL_INVALID_HANDLE, eErrors.ABC_RETVAL_INVALID_PARM
     Throw New Exception(rv.ToString)
    Case Else
     If IsError(rv) Then
      CameraError = rv
     Else
      CameraError = eErrors.ABC_RETVAL_OK
     End If
   End Select
   Return gShotParms
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonGetWritableSettings(ByVal CameraHandle As IntPtr, ByVal SettingsArray() As UInteger, ByVal SettingsArraySize As UInteger) As Integer
 End Function
 Public ReadOnly Property WritableParams As Boolean()

  Get
   Dim Length As Integer = [Enum].GetNames(GetType(eImgSetting)).Length
   Dim iWritable(Length - 1) As UInteger
   If IsNothing(gWritableParms) OrElse gWritableParms.Length <> Length Then
    ReDim gWritableParms(Length - 1)
   End If

   Dim rv As eErrors = ABCanonGetWritableSettings(CameraHandle, iWritable, Length)
   ' Returned values ABC_RETVAL_SEM_ERROR, ABC_RETVAL_INVALID_HANDLE, ABC_RETVAL_INVALID_PARM
   ' should not occur so we handle them with an exception
   Select Case rv
    Case eErrors.ABC_RETVAL_SEM_ERROR, eErrors.ABC_RETVAL_INVALID_HANDLE, eErrors.ABC_RETVAL_INVALID_PARM
     Throw New Exception(rv.ToString)
    Case Else
     If IsError(rv) Then
      CameraError = rv
     Else
      CameraError = eErrors.ABC_RETVAL_OK
     End If
   End Select
   For i As Integer = 0 To Length - 1
    gWritableParms(i) = (iWritable(i) <> 0)
   Next
   Return gWritableParms
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonCheckImgSettingWritable(ByVal CameraHandle As IntPtr, ByVal ImgSetting As UInteger) As Integer
 End Function
 Public Function IsParamWritable(ByVal ImgSetting As eImgSetting) As Boolean

  ' ABCanonCheckShotParamWritable return either ABC_RETVAL_OK or ABC_RETVAL_INVALID_PARM
  Dim rv As eErrors = ABCanonCheckImgSettingWritable(CameraHandle, ImgSetting)
  If rv = eErrors.ABC_RETVAL_OK Then
   Return True
  Else
   Return False
  End If

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 ' Basically only ABC_RETVAL_BUFFER_TOO_SMALL and ABC_RETVAL_INVALID_HANDLE are returned
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonGetList(ByVal CameraHandle As IntPtr, ByVal ShotParam As eImgSetting,
                                          ByVal ListArray() As UInteger, ByVal ListArraySize As UInteger) As Integer
 End Function
 Public ReadOnly Property ShotParamValueList(ByVal ShotParam As eImgSetting) As UInteger()

  Get
   ' We call the library function twice: first to get the array size, and a second time to retrieve the values
   ' and we do this until the values are correct (they may change between the two calls)
   Dim ListArraySize As Integer
   Dim NewListSize As Integer
   Dim ListArray() As UInteger
   Do
    ListArraySize = ABCanonGetList(CameraHandle, ShotParam, Nothing, 0)
    If IsError(ListArraySize) Then
     Return Nothing
    End If
    ReDim ListArray(ListArraySize - 1)
    NewListSize = ABCanonGetList(CameraHandle, ShotParam, ListArray, ListArraySize)
    If IsError(NewListSize) Then
     Return Nothing
    End If
   Loop While (ListArraySize <> NewListSize)
   Return ListArray
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonImgSettingValueToText(ByVal ImgSetting As eImgSetting, ByVal Value As UInteger) As IntPtr
 End Function
 Public Shared Function ShotParamToText(ByVal ShotParam As eImgSetting, ByVal Value As UInteger) As String

  Dim StrPtr As IntPtr = ABCanonImgSettingValueToText(ShotParam, Value)
  If StrPtr = Nothing Then
   Return Nothing
  End If
  Return Marshal.PtrToStringAnsi(StrPtr)

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 Public Enum eSetType As Integer
  First = 0
  Change = 1
  SetToValue = 2
  Last = 3
  PreviousValue = 4
  NextValue = 5
 End Enum
 Public Function SetImgSetting(ByVal ImgSetting As eImgSetting, ByVal SetType As eSetType, ByVal Value As Integer) As eErrors

  Dim rv As eErrors = ABCanonSetImgSetting(CameraHandle, ImgSetting, SetType, Value)
  If IsError(rv) Then
   CameraError = rv
  Else
   CameraError = eErrors.ABC_RETVAL_OK
  End If
  Return rv

 End Function
#If VBC_VER >= 11 Then
 Private Async Function iExecAsync(ByVal ct As CancellationToken, ByVal Op As Func(Of eErrors)) As Task(Of eErrors)

  Dim rv As eErrors = Op()
  If rv <> eErrors.ABC_RETVAL_BUSY Then
   If IsError(rv) Then
    CameraError = rv
   Else
    CameraError = eErrors.ABC_RETVAL_OK
   End If
   Return rv
  End If

  ' All errors except BUSY don't arrive here so...
  Do

   Dim TSC As New TaskCompletionSource(Of Object)
   ' https://stackoverflow.com/questions/18670111/task-from-cancellation-token
   ct.Register(Sub()
                TSC.TrySetCanceled()
               End Sub)
   SyncLock gTCSs
    gTCSs.Add(TSC)
   End SyncLock

   rv = Op()
   If rv = eErrors.ABC_RETVAL_BUSY Then
    Debug.Print("Async await inizio")
    '  .ConfigureAwait(False)  ' https://blog.stephencleary.com/2012/02/async-and-await.html
    Try
     Await TSC.Task.ConfigureAwait(False)
     Debug.Print("Async await fine")
     rv = Op()
    Catch
     Debug.Print("Async await catch")
     rv = eErrors.ABC_RETVAL_CANCELLED
    End Try
   End If

   SyncLock gTCSs
    Try
     Debug.Print("Async Remove(TSC)")
     gTCSs.Remove(TSC)
    Catch
    End Try
   End SyncLock

  Loop While rv = eErrors.ABC_RETVAL_BUSY

  If IsError(rv) Then
   CameraError = rv
  Else
   CameraError = eErrors.ABC_RETVAL_OK
  End If
  Return rv

 End Function
 Public Async Function GetImgSettingAsync(ByVal ImgSetting As eImgSetting) As Task(Of eErrors)

  Return Await GetImgSettingAsync(ImgSetting, CancellationToken.None)

 End Function
 Public Async Function GetImgSettingAsync(ByVal ImgSetting As eImgSetting, ByVal ct As CancellationToken) As Task(Of eErrors)

  Dim rv As eErrors = ABCanonGetImgSetting(CameraHandle, ImgSetting)
  Return Await iExecAsync(ct, Function()
                               Return ABCanonGetImgSetting(CameraHandle, ImgSetting)
                              End Function)

 End Function
 Public Async Function GetImgSettingTaskAsync(ByVal ImgSetting As eImgSetting, ByVal ct As CancellationToken) As Task(Of eErrors)

  Return Await Task.Run(Async Function()
                         Return Await GetImgSettingAsync(ImgSetting, ct)
                        End Function)

 End Function
 Public Async Function SetImgSettingAsync(ByVal ImgSetting As eImgSetting, ByVal SetType As eSetType, ByVal Value As Integer) As Task(Of eErrors)

  Return Await SetImgSettingAsync(ImgSetting, SetType, Value, CancellationToken.None)

 End Function
 Public Async Function SetImgSettingAsync(ByVal ImgSetting As eImgSetting, ByVal SetType As eSetType, ByVal Value As Integer, ByVal ct As CancellationToken) As Task(Of eErrors)

  Return Await iExecAsync(ct, Function()
                               Return ABCanonSetImgSetting(CameraHandle, ImgSetting, SetType, Value)
                              End Function)

 End Function
 Public Async Function SetImgSettingTaskAsync(ByVal ImgSetting As eImgSetting, ByVal SetType As eSetType, ByVal Value As Integer, ByVal ct As CancellationToken) As Task(Of eErrors)

  Return Await Task.Run(Async Function()
                         Return Await SetImgSettingAsync(ImgSetting, SetType, Value, ct)
                        End Function)

 End Function
 Public Async Function GetAllImgSettingsAsync(ByVal ct As CancellationToken) As Task(Of UInteger())

  Dim Length As Integer = [Enum].GetNames(GetType(eImgSetting)).Length
  Dim lShotParms(Length - 1) As UInteger

  Dim rv As eErrors = Await iExecAsync(ct, Function()
                                            Return ABCanonGetAllSettings(CameraHandle, lShotParms, Length)
                                           End Function)

  ' Returned values ABC_RETVAL_SEM_ERROR, ABC_RETVAL_INVALID_HANDLE, ABC_RETVAL_INVALID_PARM
  ' should not occur so we handle them with an exception
  Select Case rv
   Case eErrors.ABC_RETVAL_SEM_ERROR, eErrors.ABC_RETVAL_INVALID_HANDLE, eErrors.ABC_RETVAL_INVALID_PARM
    Throw New Exception(rv.ToString)
   Case Else
    Return lShotParms
  End Select

 End Function
 Public Async Function GetAllImgSettingsAsync() As Task(Of UInteger())

  Return Await GetAllImgSettingsAsync(CancellationToken.None)

 End Function
 Public Async Function GetAllImgSettingsTaskAsync(ByVal ct As CancellationToken) As Task(Of UInteger())

  Return Await Task.Run(Async Function()
                         Return Await GetAllImgSettingsAsync(ct)
                        End Function)

 End Function
 Public Async Function GetShotParamValueListAsync(ByVal ShotParam As eImgSetting, ByVal ct As CancellationToken) As Task(Of UInteger())

  Dim lParamList(-1) As UInteger

  Dim rv As eErrors = Await iExecAsync(ct, Function()
                                            Dim ListArraySize As Integer
                                            Dim NewListSize As Integer
                                            Do
                                             ListArraySize = ABCanonGetList(CameraHandle, ShotParam, Nothing, 0)
                                             If IsError(ListArraySize) Then
                                              Return ListArraySize
                                             End If
                                             ReDim lParamList(ListArraySize - 1)
                                             NewListSize = ABCanonGetList(CameraHandle, ShotParam, lParamList, ListArraySize)
                                             If IsError(NewListSize) Then
                                              Return NewListSize
                                             End If
                                            Loop While (ListArraySize <> NewListSize)
                                            Return eErrors.ABC_RETVAL_OK
                                           End Function)

  If rv <> eErrors.ABC_RETVAL_OK Then
   Throw New Exception(rv.ToString)
  End If

  Return lParamList

 End Function
#End If
 ' -------------------------------------------------------------------------------------------------------------------------
 ' Basically only ABC_RETVAL_INVALID_HANDLE is returned
 Public Enum eBatteryStatus As Integer
  ABC_BATTERY_NORMAL = 6
  ABC_BATTERY_WEAK = 4
  ABC_BATTERY_LOW = 3
  ABC_BATTERY_LB = 2
 End Enum
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetBatteryStatus(ByVal CameraHandle As IntPtr) As Integer
 End Function
 Public Function BatteryStatus() As Integer

  Return ABCanonGetBatteryStatus(CameraHandle)

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 Private Delegate Function ABC_NOTIFICATION_CB(ByVal CameraHandle As IntPtr, ByRef ABCEventInternal As sEventInternal) As Integer
 Private MyNotificationFunc As ABC_NOTIFICATION_CB = AddressOf NotificationFunction

 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Sub ABCanonReleaseEvent(ByVal pEvent As IntPtr)
 End Sub

 ' Con la versione Task.Run, quando termina il busy mi crasha con l'eccezione:
 ' An unhandled exception of type 'System.InvalidOperationException' occurred in mscorlib.dll
 ' Additional information: La raccolta è stata modificata. L'operazione di enumerazione potrebbe non essere eseguita.
 ' Nel Next di foreach.
 ' Questo perche' con 2 task girano assieme e mentre uno continua e rimuove il coso dalla lista come se non fosse stato fatto il
 ' setresult, l'altro continua e si accorge che la lista e' stata modificata
 ' Il che vuol dire che synclock non serializza un tubo. In realtà no: la SetResult viene eseguita sincrona!

 ' Questa .SetResult è una rogna. Per default, viene eseguita sincrona, quindi da qui continua DIRETTAMENTE nel codice dove si è fatto l'await 
 ' Il che, volendo gestire la gTSCs come lista con un SemaphoreSlim mi deadlocka di sicuro.
 ' Se non uso .ConfigureAwait(false) probabilmente mi funziona se la funzione originale è chiamata dalla UI, dal momento che questa NON è di sicuro
 ' chiamata dalla UIThread e quindi non lo può fare. Ma cmq non mi funzionerà se uso la versione "AsyncAsync" con il Task.Run (perchè a questo punto
 ' mi aspetto che questa thread sia equivalente alla thread creata dalla Task.Run).
 ' La cosa più semplice sarebbe avere TaskCreationOptions.RunContinuationsAsynchronously nella TaskCompletionSource ma questo compare con NET 4.6.
 ' Oppure mettere i .SetResult in una serie di Task.Run separati in modo che QUESTA routine finisca e rilasci il SemaphoreSlim.
 '
 ' Un'idea potrebbe essere quella di trasformare questo foreach in un  while(.Count>0) {X=Elemento[0], Remove(X), X.SetResult}. Il SyncLock continuerebbe
 ' a garantirmi la coerenza se le thread sono più di una, ma la sincronicità di SetResult() mi sta rompendo SOLO per il fatto che schianta il foreach.
 ' Poi magari mi serve un Task.Run() per non far girare troppa roba sulla NotificationFunction che probabilmente mi blocca la thread nativa per troppo
 ' tempo, ma questa è da profilare, prima.

 Private Function NotificationFunction(ByVal CameraHandle As IntPtr, ByRef ABCEventInternal As sEventInternal) As Integer

  Dim ABCEvent As sEvent

  Select Case ABCEventInternal.EventID
   Case eEventID.ABC_EVENT_BUSY
    If ABCEventInternal.EventSize = 0 Then   ' End of Busy
     If CameraError = eErrors.ABC_RETVAL_BUSY Then
      CameraError = eErrors.ABC_RETVAL_OK
     End If

#If VBC_VER >= 11 Then
     Dim TCS As TaskCompletionSource(Of Object)
     Do
      SyncLock gTCSs
       If gTCSs.Count > 0 Then
        TCS = gTCSs.Item(0)
        gTCSs.Remove(TCS)
        Debug.Print("TCS Found and removed")
       Else
        TCS = Nothing
        Debug.Print("No TCS")
       End If
      End SyncLock
      If TCS IsNot Nothing Then
       Try
        Debug.Print("TCS.SetResult begin")
        TCS.SetResult(Nothing)
        Debug.Print("TCS.SetResult end")
       Catch
       End Try
      End If
     Loop While TCS IsNot Nothing
    End If
#End If

   Case eEventID.ABC_EVENT_DOWNLOAD_COMPLETED
    If gShootTCS IsNot Nothing Then
     Try
      gShootTCS.SetResult(Nothing)
     Catch ex As Exception
     End Try
    End If
  End Select

  ABCEvent.EventID = ABCEventInternal.EventID
  ABCEvent.EventSize = ABCEventInternal.EventSize
  ABCEvent.EventPtr = ABCEventInternal.EventPtr
  ABCEvent.Handled = False

  RaiseEvent CameraEvent(Me, ABCEvent)

  If ABCEvent.Handled Then
   Return 1
  Else
   Return 0
  End If

 End Function
 Public Shared Function GetTextForEvent(ByVal Evt As sEvent) As String

  Select Case Evt.EventID
   Case eEventID.ABC_EVENT_JPG_SAVED, eEventID.ABC_EVENT_CR2_SAVED
    Return Marshal.PtrToStringUni(Evt.EventPtr)
   Case Else
    Return Nothing
  End Select

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonSetImageCounter(ByVal CameraHandle As IntPtr, ByVal Counter As UInt32) As Integer
 End Function
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonGetImageCounter(ByVal CameraHandle As IntPtr, ByRef pCounter As UInt32) As Integer
 End Function
 Public Property ImageCounter As UInt32

  Get
   Dim rv As UInt32
   ABCanonGetImageCounter(CameraHandle, rv)
   Return rv
  End Get
  Set(value As UInt32)
   ABCanonSetImageCounter(CameraHandle, value)
  End Set

 End Property
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Unicode)> _
 Private Shared Function ABCanonSetSaveFilenameW(ByVal CameraHandle As IntPtr, ByVal FileName As String, ByVal DigitsCount As Integer) As Integer
 End Function
 Public Function SetSaveFilename(Filename As String, ByVal DigitsCount As Integer) As Boolean

  If ABCanonSetSaveFilenameW(CameraHandle, Filename, DigitsCount) = eErrors.ABC_RETVAL_OK Then
   Return True
  Else
   Return False
  End If

 End Function
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Unicode)> _
 Private Shared Function ABCanonSetSaveOverrideW(ByVal CameraHandle As IntPtr, ByVal FileName As String) As Integer
 End Function
 Public Function SetOverrideFilename(Filename As String) As Boolean

  If ABCanonSetSaveOverrideW(CameraHandle, Filename) = eErrors.ABC_RETVAL_OK Then
   Return True
  Else
   Return False
  End If

 End Function
 Public Enum eDownloadMode As Integer
  Discard = 0
  SaveToFile = 1
  ToMemory = 2
 End Enum
 ' Basically only ABC_RETVAL_INVALID_PARM and ABC_RETVAL_INVALID_HANDLE are returned
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonSetDownloadMode(ByVal CameraHandle As IntPtr, ByVal DownloadMode As eDownloadMode) As Integer
 End Function
 Public Sub SetDownloadMode(ByVal DownloadMode As eDownloadMode)

  Dim rv As eErrors = ABCanonSetDownloadMode(CameraHandle, DownloadMode)
  If IsError(rv) Then
   Throw New Exception(rv.ToString)
  End If

 End Sub
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonSetSaveTarget(ByVal CameraHandle As IntPtr, ByVal SaveTarget As Integer) As Integer
 End Function
 Public Enum eSaveTarget As Integer
  SaveToPC = 2
  SaveToCF = 8
  SaveToBoth = 10
 End Enum
 Public Function SaveTo(ByVal SaveTarget As eSaveTarget) As eErrors

  Dim rv As eErrors = ABCanonSetSaveTarget(CameraHandle, SaveTarget)
  If IsError(rv) Then
   CameraError = rv
  Else
   CameraError = eErrors.ABC_RETVAL_OK
  End If
  Return rv

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 Public Enum eShootingDeniedReason As Integer
  ALLOWED = 0
  MIRROR_LOCK = 1
  SELF_TIMER = 2
  BULB = 3
 End Enum
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonIsShootingAllowed(ByVal CameraHandle As IntPtr, ByRef Reason As Integer) As Integer
 End Function
 Public ReadOnly Property IsShootAllowed As eShootingDeniedReason

  Get
   Dim Reason As eShootingDeniedReason
   Dim rv As eErrors = ABCanonIsShootingAllowed(CameraHandle, Reason)
   Select Case rv
    Case eErrors.ABC_RETVAL_DENIED, eErrors.ABC_RETVAL_OK
     ' Good responses, return reason
     Return Reason
    Case Else
     ' Unexpected responses (should not occurs), throw exception
     Throw New Exception(rv.ToString)
   End Select
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonShoot(ByVal CameraHandle As IntPtr) As Integer
 End Function
 Public Function Shoot() As eErrors

  Dim rv As eErrors = ABCanonShoot(CameraHandle)
  If IsError(rv) Then
   CameraError = rv
  Else
   CameraError = eErrors.ABC_RETVAL_OK
  End If
  Return rv

 End Function
#If VBC_VER >= 11 Then
 Public Async Function ShootAsyc() As Task(Of eErrors)

  If gShootTCS IsNot Nothing Then
   Throw New InvalidOperationException("Shooting already in progress")
  End If

  gShootTCS = New TaskCompletionSource(Of Object)

  Dim rv As eErrors = ABCanonShoot(CameraHandle)

  If IsError(rv) Then
   CameraError = rv
  Else
   CameraError = eErrors.ABC_RETVAL_OK
   Await gShootTCS.Task
  End If

  gShootTCS = Nothing
  Return rv

 End Function
#End If

 ' -------------------------------------------------------------------------------------------------------------------------
 Public Enum eOptions As UInt32
  BULB_IN_LIST = 1
  HANDLE_THREAD_BUSY = 2
  STANDARD_BULB = 4
  ALLOW_OVERWRITE = 8
  PERMANENT_OVERRIDE = 16
  BUSYPOLL = 32
  PCT2 = 64
 End Enum
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonSetOptions(ByVal CameraHandle As IntPtr, ByVal Options As eOptions) As Integer
 End Function
 <DllImport(DLL, SetLastError:=False)> _
 Private Shared Function ABCanonGetOptions(ByVal CameraHandle As IntPtr, ByRef Options As eOptions) As Integer
 End Function
 Public Property Options As eOptions

  Get
   Dim Opt As eOptions
   ABCanonGetOptions(CameraHandle, Opt)
   Return Opt
  End Get
  Set(value As eOptions)
   ABCanonSetOptions(CameraHandle, value)
  End Set

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetRemainingShots(ByVal CameraHandle As IntPtr) As Integer
 End Function
 Public Function RemainingShots() As Integer

  Return ABCanonGetRemainingShots(CameraHandle)

 End Function
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetCFStatus(ByVal CameraHandle As IntPtr, ByRef TotalKB As UInt32, ByRef FreeKB As UInt32) As Integer
 End Function
 Public ReadOnly Property CFFreeKbytes() As UInt32

  Get
   Dim TotalKB As UInt32
   Dim FreeKB As UInt32
   Dim rv As Integer = ABCanonGetCFStatus(CameraHandle, TotalKB, FreeKB)
   If IsError(rv) Then
    Return 0
   End If
   Return FreeKB
  End Get

 End Property
 Public ReadOnly Property CFTotalKbytes() As UInt32

  Get
   Dim TotalKB As UInt32
   Dim FreeKB As UInt32
   Dim rv As Integer = ABCanonGetCFStatus(CameraHandle, TotalKB, FreeKB)
   If IsError(rv) Then
    Return 0
   End If
   Return TotalKB
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 Public Enum eCustomFunction As Integer
  ABC_CF_SETBUTTON_FUNCT = &HAB0000
  ABC_CF_LONGEXP_NOISE_RED = &HAB0001
  ABC_CF_FLASH_SYNC = &HAB0002
  ABC_CF_SHUTTER_AE_LOCK_BTN = &HAB0003
  ABC_CF_AF_ASSIST_BEAM = &HAB0004
  ABC_CF_EXP_LEVEL_INCREMENT = &HAB0005
  ABC_CF_FLASH_FIRING = &HAB0006
  ABC_CF_ISO_EXPANSION = &HAB0007
  ABC_CF_AEB_SEQUENCE = &HAB0008
  ABC_CF_SUPERIMPOSED_DISPLAY = &HAB0009
  ABC_CF_MENU_BUTTON_DISPLAY = &HAB000A
  ABC_CF_MIRROR_LOCKUP = &HAB000B
  ABC_CF_AF_POINT_SELECTION = &HAB000C
  ABC_CF_E_TTL_II = &HAB000D
  ABC_CF_SHUTTER_CURTAIN = &HAB000E
  ABC_CF_SAFETY_SHIFT = &HAB000F
  ABC_CF_AF_POINT_ACTIVATION = &HAB0010
  ABC_CF_RETURN_TO_SHOOT = &HAB0011
  ABC_CF_LENS_AF_STOP_BUTTON = &HAB0012
  ABC_CF_ADD_ORIGINAL_DATA = &HAB0013
 End Enum
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetCustomFunction(ByVal CameraHandle As IntPtr, ByVal CustomOption As eCustomFunction) As Integer
 End Function
 Public ReadOnly Property CustomFunction(ByVal Idx As eCustomFunction) As Integer

  Get
   Dim rv As Integer = ABCanonGetCustomFunction(CameraHandle, Idx)
   If IsError(rv) Then
    CameraError = rv
    rv = 0
   Else
    CameraError = eErrors.ABC_RETVAL_OK
   End If

   Return rv
  End Get

 End Property
 ' -------------------------------------------------------------------------------------------------------------------------
 <DllImport(DLL, SetLastError:=False, CharSet:=CharSet.Ansi)> _
 Private Shared Function ABCanonGetPQImageCount(ByVal CameraHandle As IntPtr, ByVal PictureQuality As eSPPictureQuality) As Integer
 End Function
 Public ReadOnly Property PQImageCount As Integer

  Get
   Dim rv As Integer = ABCanonGetPQImageCount(CameraHandle, -1)
   If IsError(rv) Then
    Return 0
   End If
   Return rv
  End Get

 End Property
End Class
